


/*********************************************************************
 * (C) Copyright 2010 INRIA, France
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 * 
 *********************************************************************/


/*
 * THIS SOURCE CODE IS SUPPLIED  ``AS IS'' WITHOUT WARRANTY OF ANY KIND, 
 * AND ITS AUTHOR AND THE JOURNAL OF ARTIFICIAL INTELLIGENCE RESEARCH 
 * (JAIR) AND JAIR'S PUBLISHERS AND DISTRIBUTORS, DISCLAIM ANY AND ALL 
 * WARRANTIES, INCLUDING BUT NOT LIMITED TO ANY IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE, AND
 * ANY WARRANTIES OR NON INFRINGEMENT.  THE USER ASSUMES ALL LIABILITY AND
 * RESPONSIBILITY FOR USE OF THIS SOURCE CODE, AND NEITHER THE AUTHOR NOR
 * JAIR, NOR JAIR'S PUBLISHERS AND DISTRIBUTORS, WILL BE LIABLE FOR 
 * DAMAGES OF ANY KIND RESULTING FROM ITS USE.  Without limiting the 
 * generality of the foregoing, neither the author, nor JAIR, nor JAIR's
 * publishers and distributors, warrant that the Source Code will be 
 * error-free, will operate without interruption, or will meet the needs 
 * of the user.
 */




/*********************************************************************
 * File: main.c
 * Description: The main routine for FF-based TorchLight
 *
 * Author: Joerg Hoffmann 2010
 * 
 *********************************************************************/ 








#include "torchlight.h"

#include "memory.h"
#include "output.h"

#include "parse.h"

#include "inst_pre.h"
#include "inst_easy.h"
#include "inst_hard.h"
#include "inst_final.h"

#include "relax.h"
#include "search.h"

#include "SG-DTGs.h"
#include "gDGs-lDGs.h"
#include "oDGplus.h"
#include "lEHC.h"


















/*
 *  ----------------------------- GLOBAL VARIABLES ----------------------------
 */












/*******************
 * GENERAL HELPERS *
 *******************/








/* used to time the different stages of the planner
 */
float gtempl_time = 0, greach_time = 0, grelev_time = 0, gconn_time = 0;
float gsearch_time = 0, gSG_DTG_time = 0, gSG_DTG_static_analyze_time = 0, ganalyze_global_time = 0;
float ganalyze_local_ini_lDG_time = 0, ganalyze_local_ini_oDGplus_time = 0;
float ganalyze_local_sample_lDG_time = 0, ganalyze_local_sample_oDGplus_time = 0;
float gsampling_time = 0, grelaxed_plan_time = 0, gFDvariables_time = 0;
float ganalyze_local_sample_monotoneEHC_time = 0;


/* the command line inputs
 */
struct _command_line gcmd_line;

/* number of states that got heuristically evaluated
 */
int gevaluated_states = 0;

/* maximal depth of breadth first search
 */
int gmax_search_depth = 0;





/***********
 * PARSING *
 ***********/





/* resulting name for ops file
 */
char gops_file[MAX_LENGTH] = "";
/* same for fct file 
 */
char gfct_file[MAX_LENGTH] = "";


/* used for pddl parsing, flex only allows global variables
 */
int gbracket_count;
char *gproblem_name;

/* The current input line number
 */
int lineno = 1;

/* The current input filename
 */
char *gact_filename;

/* The pddl domain name
 */
char *gdomain_name = NULL;

/* loaded, uninstantiated operators
 */
PlOperator *gloaded_ops = NULL;

/* stores initials as fact_list 
 */
PlNode *gorig_initial_facts = NULL;

/* not yet preprocessed goal facts
 */
PlNode *gorig_goal_facts = NULL;

/* axioms as in UCPOP before being changed to ops
 */
PlOperator *gloaded_axioms = NULL;

/* the types, as defined in the domain file
 */
TypedList *gparse_types = NULL;

/* the constants, as defined in domain file
 */
TypedList *gparse_constants = NULL;

/* the predicates and their arg types, as defined in the domain file
 */
TypedListList *gparse_predicates = NULL;

/* the objects, declared in the problem file
 */
TypedList *gparse_objects = NULL;


/* connection to instantiation ( except ops, goal, initial )
 */

/* all typed objects 
 */
FactList *gorig_constant_list = NULL;

/* the predicates and their types
 */
FactList *gpredicates_and_types = NULL;












/*****************
 * INSTANTIATING *
 *****************/









/* global arrays of constant names,
 *               type names (with their constants),
 *               predicate names,
 *               predicate aritys,
 *               defined types of predicate args
 */
Token gconstants[MAX_CONSTANTS];
int gnum_constants = 0;
Token gtype_names[MAX_TYPES];
int gtype_consts[MAX_TYPES][MAX_TYPE];
Bool gis_member[MAX_CONSTANTS][MAX_TYPES];
int gmember_nr[MAX_CONSTANTS][MAX_TYPES];/* nr of object within a type */
int gtype_size[MAX_TYPES];
int gnum_types = 0;
Token gpredicates[MAX_PREDICATES];
int garity[MAX_PREDICATES];
int gpredicates_args_type[MAX_PREDICATES][MAX_ARITY];
int gnum_predicates = 0;





/* the domain in integer (Fact) representation
 */
Operator_pointer goperators[MAX_OPERATORS];
int gnum_operators = 0;
Fact *gfull_initial;
int gnum_full_initial = 0;
WffNode *ggoal = NULL;




/* stores inertia - information: is any occurence of the predicate
 * added / deleted in the uninstantiated ops ?
 */
Bool gis_added[MAX_PREDICATES];
Bool gis_deleted[MAX_PREDICATES];



/* splitted initial state:
 * initial non static facts,
 * initial static facts, divided into predicates
 * (will be two dimensional array, allocated directly before need)
 */
Facts *ginitial = NULL;
int gnum_initial = 0;
Fact **ginitial_predicate;
int *gnum_initial_predicate;



/* the type numbers corresponding to any unary inertia
 */
int gtype_to_predicate[MAX_PREDICATES];
int gpredicate_to_type[MAX_TYPES];

/* (ordered) numbers of types that new type is intersection of
 */
TypeArray gintersected_types[MAX_TYPES];
int gnum_intersected_types[MAX_TYPES];



/* splitted domain: hard n easy ops
 */
Operator_pointer *ghard_operators;
int gnum_hard_operators;
NormOperator_pointer *geasy_operators;
int gnum_easy_operators;



/* so called Templates for easy ops: possible inertia constrained
 * instantiation constants
 */
EasyTemplate *geasy_templates;
int gnum_easy_templates;



/* first step for hard ops: create mixed operators, with conjunctive
 * precondition and arbitrary effects
 */
MixedOperator *ghard_mixed_operators;
int gnum_hard_mixed_operators;



/* hard ''templates'' : pseudo actions
 */
PseudoAction_pointer *ghard_templates;
int gnum_hard_templates;



/* store the final "relevant facts"
 */
Fact grelevant_facts[MAX_RELEVANT_FACTS];
int gnum_relevant_facts = 0;
int gnum_pp_facts = 0;



/* the final actions and problem representation
 */
Action *gactions;
int gnum_actions;
State ginitial_state;
State ggoal_state;









/**********************
 * CONNECTIVITY GRAPH *
 **********************/







/* one ops (actions) array ...
 */
OpConn *gop_conn;
int gnum_op_conn;



/* one effects array ...
 */
EfConn *gef_conn;
int gnum_ef_conn;



/* one facts array.
 */
FtConn *gft_conn;
int gnum_ft_conn;









/*******************
 * SEARCHING NEEDS *
 *******************/








/* the goal state, divided into subsets
 */
State *ggoal_agenda;
int gnum_goal_agenda;



/* byproduct of fixpoint: applicable actions
 */
int *gA;
int gnum_A;



/* communication from extract 1.P. to search engines:
 * 1P action choice
 */
int *gH;
int gnum_H;



/* the effects that are considered true in relaxed plan
 */
int *gin_plan_E;
int gnum_in_plan_E;



/* always stores (current) serial plan
 */
int gplan_ops[MAX_PLAN_LENGTH];
int gnum_plan_ops = 0;



/* stores the states that the current plan goes through
 * ( for knowing where new agenda entry starts from )
 */
State gplan_states[MAX_PLAN_LENGTH + 1];










/************************************
 * GLOBAL VARS FOR CG-->H+ ANALYSIS *
 ************************************/




/* the vars, parsed from FD
 */
Variable *gvariables;
int gnum_variables;



/* the support graph
 */
SG gSG;



/* the DTG for each var
 */
DTG *gDTGs;



/* just a helper...
 */
int *ggoal_on_var;



/* must communicate relaxed plan to oDGplus analysis.
 */
int *grplan;
int gnum_rplan;



/* to communicate back analysis results
 */
Bool gsuccess;
int ged_bound;
Bool gdead_end;

float gsuccess_percentage;
int gmin_ed_bound;
float gmean_ed_bound;
int gmax_ed_bound;
float gdead_end_percentage;



/* the sample states.
 */
State *gsample_states;



/* the data output file, if wanted.
 */
FILE *goutput_file;



/* for diagnosis!
 *
 * First of all, a flag whether we're actually diagnosing the reasons
 * for negative outcome -- on option, we want to do this diagnosis
 * only for those states for which actually had this outcome. The
 * outcome will be known only posthum so have to run analysis two
 * times.
 */
Bool gdo_negative_diagnosis;

/* counts of how many successful/failed analysis attempts
 */
int ggDG_num_graphs;/* for gDG, also need to count number of graphs considered */
int ggDG_num_successes;
int glDG_num_successes; /* here num trials is simply number of samples */
int goDGplus_num_successes;
int goDGplus_num_failures;

/* The first guy here will be the union of success var0s in gDG
 * analysis.
 */
int *ggDG_responsible_var0s;
int *ggDG_responsible_op0s;
int *ggDG_responsible_op0var0s_weights;/* how many times this guy? */
int ggDG_num_responsible_op0var0s;

/* The next guy will be the union of success var0s in lDG analysis.
 */
int *glDG_responsible_var0s;
int *glDG_responsible_var0s_weights;/* how many times this guy? */
int glDG_num_responsible_var0s;

/* These here will be the pairs of (successful t0-eff-pred, op0-actionname) for oDG+
 * analysis.
 */
int *goDGplus_responsible_pred0s;
int *goDGplus_responsible_op0s;/* index into goperators array! */
int *goDGplus_responsible_op0pred0s_weights;
int goDGplus_num_responsible_op0pred0s;

/* Here are the variables that appeared on a cycle in oDG+ analysis.
 */
int *goDGplus_cycle_vars;
int *goDGplus_cycle_vars_weights;
int goDGplus_num_cycle_vars;

/* /\* Here are all facts that were in non-recovered RFC_intersect */
/*  *\/ */
/* int *goDGplus_nonrecovered_RFC_intersect; */
/* int goDGplus_num_nonrecovered_RFC_intersect; */

/* Instead, collect pair of op0-actionname and ft-pred of RFS Intersect.
 */
int *goDGplus_nonrecovered_RFC_intersect_preds;
int *goDGplus_nonrecovered_RFC_intersect_op0s;/* index into goperators array! */
int *goDGplus_nonrecovered_RFC_intersects_weights;
int goDGplus_num_nonrecovered_RFC_intersects;
int goDGplus_nonrecovered_RFC_intersects_totalweight;

/* for comparison, here's the alternative previously used, recording
 * op0/var instead.
 */
int *gOLD_oDGplus_nonrecovered_RFC_intersect_vars;
int *gOLD_oDGplus_nonrecovered_RFC_intersect_op0s;/* index into goperators array! */
int *gOLD_oDGplus_nonrecovered_RFC_intersects_weights;
int gOLD_oDGplus_num_nonrecovered_RFC_intersects;
int gOLD_oDGplus_nonrecovered_RFC_intersects_totalweight = 0;

/* For any non-leaf transition that was invertible/induced but had
 * side effects, here is the pair of responsible rop-actionname and
 * side-effect variable.
 */
int *goDGplus_nonleafbadtrans_seffvars;
int *goDGplus_nonleafbadtrans_rops;/* index into goperators array! */
int *goDGplus_nonleafbadtranss_weights;
int goDGplus_num_nonleafbadtranss;

/* flag so that SG cycle check knows who he is working for.
 * 1 = gDG, 2 = lDG, 3 = oDG+
 */
int gchecking_acyclic_for;



/* for EHC run, checking the "predictive quality" of the sampling.
 */
int gEHC_success;
int gEHC_max_search_depth;
float gEHC_time = 0;
Bool doing_EHC = FALSE; /* flag indicating that we're now no longer in analysis mode */ 



/* signal to EHC that we're not actually doing search but local-EHC
 * analysis
 */
Bool gdoing_EHCanalyze = FALSE;






















/*
 *  ----------------------------- HEADERS FOR PARSING ----------------------------
 * ( fns defined in the scan-* files )
 */







void get_fct_file_name( char *filename );
void load_ops_file( char *filename );
void load_fct_file( char *filename );











/*
 *  ----------------------------- MAIN ROUTINE ----------------------------
 */











int main( int argc, char *argv[] )

{

  struct tms start, end;
  int i, j, op;

  int maxw;
  int maxi;
  Bool gotone;
  float oDGsuccessrate;



  /* command line treatment
   */
  gcmd_line.num_samples = 10; /* default: 10 */
  gcmd_line.depth_samples = 5; /* default: 5 */

  gcmd_line.do_gDG = TRUE;
  gcmd_line.do_lDG = TRUE;
  gcmd_line.do_SG_root = TRUE;
  gcmd_line.optimize_over_op0var0 = FALSE;
  gcmd_line.prune_useless_var0 = FALSE;
  gcmd_line.replacement_level = 2;
  gcmd_line.do_recoverer_only_relevant = TRUE;

  gcmd_line.do_diagnose = FALSE; /* default: FALSE */
  gcmd_line.do_diagnose_gDG_successrate = TRUE;
  gcmd_line.do_diagnose_invertop0 = TRUE;
  gcmd_line.do_diagnose_ignore_exchangedvars = TRUE;
  gcmd_line.negative_diagnose_all = TRUE;
  gcmd_line.diagnose_prune_useless_var0 = TRUE;
  gcmd_line.show_case_weight = TRUE;

  gcmd_line.run_FF = FALSE;
  gcmd_line.run_only_FF = FALSE;
  gcmd_line.run_EHCanalyze = FALSE;
  gcmd_line.run_only_EHCanalyze = FALSE;

  gcmd_line.output_file = FALSE;
  gcmd_line.display_info = 1;
  gcmd_line.debug = 0;

/*   gcmd_line.do_replace_op0_to_reduce_R; */
/*   gcmd_line.do_replace_P0plus_to_reduce_R; */
/*   gcmd_line.do_replace_for_recoverRFC; */
        
  memset(gcmd_line.ops_file_name, 0, MAX_LENGTH);
  memset(gcmd_line.fct_file_name, 0, MAX_LENGTH);
  memset(gcmd_line.path, 0, MAX_LENGTH);

  if ( argc == 1 || ( argc == 2 && *++argv[0] == '?' ) ) {
    ff_usage();
    exit( 1 );
  }
  if ( !process_command_line( argc, argv ) ) {
    ff_usage();
    exit( 1 );
  }


  /* make file names
   */

  /* one input name missing
   */
  if ( !gcmd_line.ops_file_name || 
       !gcmd_line.fct_file_name ) {
    fprintf(stdout, "\nTorchLight: two input files needed\n\n");
    ff_usage();      
    exit( 1 );
  }
  /* add path info, complete file names will be stored in
   * ops_file and fct_file 
   */
  sprintf(gops_file, "%s%s", gcmd_line.path, gcmd_line.ops_file_name);
  sprintf(gfct_file, "%s%s", gcmd_line.path, gcmd_line.fct_file_name);


  /* parse the input files
   */

  /* start parse & instantiation timing
   */
  times( &start );
  /* domain file (ops)
   */
  if ( gcmd_line.display_info >= 1 ) {
    printf("\nTorchLight: parsing domain file");
  } 
  /* it is important for the pddl language to define the domain before 
   * reading the problem 
   */
  load_ops_file( gops_file );
  /* problem file (facts)
   */  
  if ( gcmd_line.display_info >= 1 ) {
    printf(" ... done.\nTorchLight: parsing problem file"); 
  }
  load_fct_file( gfct_file );
  if ( gcmd_line.display_info >= 1 ) {
    printf(" ... done.\n\n");
  }

  /* This is needed to get all types.
   */
  build_orig_constant_list();

  /* last step of parsing: see if it's an ADL domain!
   */
  if ( !make_adl_domain() ) {
    printf("\nTorchLight: this is not a STRIPS problem!");
    printf("\n            can't be handled by this version.\n\n");
    exit( 1 );
  }


  /* now instantiate operators;
   */


  /**************************
   * first do PREPROCESSING * 
   **************************/


  /* start by collecting all strings and thereby encoding 
   * the domain in integers.
   */
  encode_domain_in_integers();

  /* inertia preprocessing, first step:
   *   - collect inertia information
   *   - split initial state into
   *        _ arrays for individual predicates
   *        - arrays for all static relations
   *        - array containing non - static relations
   */
  do_inertia_preprocessing_step_1();

  /* normalize all PL1 formulae in domain description:
   * (goal, preconds and effect conditions)
   *   - simplify formula
   *   - expand quantifiers
   *   - NOTs down
   */
  normalize_all_wffs();

  /* translate negative preconds: introduce symmetric new predicate
   * NOT-p(..) (e.g., not-in(?ob) in briefcaseworld)
   */
  translate_negative_preconds();

  /* split domain in easy (disjunction of conjunctive preconds)
   * and hard (non DNF preconds) part, to apply 
   * different instantiation algorithms
   */
  split_domain();


  /***********************************************
   * PREPROCESSING FINISHED                      *
   *                                             *
   * NOW MULTIPLY PARAMETERS IN EFFECTIVE MANNER *
   ***********************************************/

  build_easy_action_templates();
  build_hard_action_templates();

  times( &end );
  TIME( gtempl_time );

  times( &start );

  /* perform reachability analysis in terms of relaxed 
   * fixpoint
   */
  perform_reachability_analysis();

  times( &end );
  TIME( greach_time );

  times( &start );

  /* collect the relevant facts and build final domain
   * and problem representations.
   */
  collect_relevant_facts();

  times( &end );
  TIME( grelev_time );

  times( &start );

  /* now build globally accessable connectivity graph
   */
  build_connectivity_graph();

  /*  will use heuristic hence need to run this step.
   */
  initialize_relax();

  times( &end );
  TIME( gconn_time );
  



  /***********************************************************
   * SG-->H+ ANALYSIS! 
   ***********************************************************/



  if ( gcmd_line.run_only_FF ) {
    if ( gcmd_line.display_info >= 1 ) {
      printf("\n\n\nTorchLight: running Enforced Hill-Climbing... "); 
      fflush(stdout);
    }
    times( &start );
    
    for ( i = 0; i < MAX_PLAN_LENGTH + 1; i++ ) {
      make_state( &(gplan_states[i]), gnum_ft_conn );
      gplan_states[i].max_F = gnum_ft_conn;
    }
    
    gEHC_max_search_depth = -1;
    doing_EHC = TRUE;
    
    if ( !do_enforced_hill_climbing( &ginitial_state, &ggoal_state ) ) {
      gEHC_success = 0;
    } else {
      gEHC_success = 1;
    }

    times( &end );
    TIME( gEHC_time );

    if ( gcmd_line.output_file ) {
      if ( (goutput_file = fopen( "EHC.txt", "w")) == NULL ) {
	printf("\nCan't open EHC.txt output file!\n\n");
	exit(1);
      }
      fprintf(goutput_file, "!!!!!-----------------------------------------------------------------\n");
      fprintf(goutput_file, "Input domain             : %s\n", gops_file);
      fprintf(goutput_file, "Input problem instance   : %s\n", gfct_file);
      fprintf(goutput_file, "----------------------------------------------------------------------\n");
      fprintf(goutput_file, "EHC runtime              : %.2f\n", gEHC_time);
      fprintf(goutput_file, "EHC success              : %d\n", gEHC_success);
      fprintf(goutput_file, "EHC max search depth     : %d\n", gEHC_max_search_depth);
      fprintf(goutput_file, "----------------------------------------------------------------------\n");
      fclose(goutput_file);
    }

    output_planner_info();

    printf("\n\n");
    exit( 0 );
   
  }


  
  if ( gcmd_line.run_only_EHCanalyze ) {
    times( &start );
    
    sample_states();
    
    times( &end );
    TIME( gsampling_time );

    if ( gcmd_line.display_info >= 1 ) {
      printf("\n\n\nTorchLight: running local-monotone Enforced Hill-Climbing analysis... "); 
      fflush(stdout);
    }
    
    times( &start );
    
    analyze_samples_local_monotoneEHC();
    
    times( &end );
    TIME( ganalyze_local_sample_monotoneEHC_time );
    if ( gcmd_line.output_file ) {
      if ( (goutput_file = fopen( "EHCsample.txt", "w")) == NULL ) {
	printf("\nCan't open EHCsample.txt output file!\n\n");
	exit(1);
      }
      fprintf(goutput_file, "!!!!!-----------------------------------------------------------------\n");
      fprintf(goutput_file, "Input domain             : %s\n", gops_file);
      fprintf(goutput_file, "Input problem instance   : %s\n", gfct_file);
      fprintf(goutput_file, "----------------------------------------------------------------------\n");
      fprintf(goutput_file, "Sample states runtime    : %.2f\n", gsampling_time);
      fprintf(goutput_file, "----------------------------------------------------------------------\n");
      fclose(goutput_file);
    }

    
    printf("\n\nTorchLight (approximate) monotone-local-EHC analysis of sampled states:");
    printf("\nSuccess and hence no local minima under h+: %6.2f%%", 
	   gsuccess_percentage);
    printf("\nDead-end states: %6.2f%%", gdead_end_percentage);
    printf("\nExit distance bound min: %4d, mean: %6.2f, max: %4d",
	   gmin_ed_bound, gmean_ed_bound, gmax_ed_bound);
    fflush(stdout);
    
    if ( gcmd_line.output_file ) {
      if ( (goutput_file = fopen( "EHCsample.txt", "a")) == NULL ) {
	printf("\nCan't open EHCsample.txt output file!\n\n");
	exit(1);
      }
      fprintf(goutput_file, "Local-monotoneEHC(sample): 1, %6.2f, %6.2f, %6.2f, %3d, %6.2f, %3d\n", 
	      ganalyze_local_sample_monotoneEHC_time,
	      gsuccess_percentage,
	      gdead_end_percentage,
	      gsuccess_percentage > 0 ? gmin_ed_bound : -1,
	      gsuccess_percentage > 0 ? gmean_ed_bound : -1,
	      gsuccess_percentage > 0 ? gmax_ed_bound : -1);
      fprintf(goutput_file, "----------------------------------------------------------------------\n");
      fclose(goutput_file);
    }

    output_planner_info();

    printf("\n\n");
    exit( 0 );
   
  }




  /* first of all, let downward's preprocessor generate the multi-valued vars,
   * and parse them!
   */
  if ( gcmd_line.display_info >= 1 ) {
    printf("\nTorchLight: running Fast-Downward translator to generate variables"); 
    fflush(stdout);
  }
  times( &start );
  
  create_and_parse_variables();
  
  times( &end );
  TIME( gFDvariables_time );
  if ( gcmd_line.display_info >= 1 ) {
    printf(" ... done.");
    fflush(stdout);
  }



  if ( gcmd_line.display_info >= 1 ) {
    printf("\nTorchLight: creating SG and DTG structures"); 
    fflush(stdout);
  }
  times( &start );
  
  /* now create the varval linking in fts, and the pre/eff in ops!
   */
  create_ft_op_indices();
  
  /* now create the support graph and the DTGs!
   */
  create_SG_DTG();
  
  if ( gcmd_line.display_info >= 1 ) {
    printf(" ... done.");
    fflush(stdout);
  }
  
  times( &end );
  TIME( gSG_DTG_time );





  if ( gcmd_line.display_info >= 1 ) {
    printf("\nTorchLight: static examination of SG and DTG structures"); 
    fflush(stdout);
  }
  times( &start );
  
  /* determine SG and DTG static properties -- acylic, t invertible, etc.
   */
  determine_SG_DTG_static_properties();
  
  if ( gcmd_line.display_info >= 1 ) {
    printf(" ... done.\n\n");
    fflush(stdout);
  }
  
  times( &end );
  TIME( gSG_DTG_static_analyze_time );
  
  if ( gcmd_line.display_info > 1 ) {
    printf("\n\n========================VARIABLES:\n");
    for ( i = 0; i < gnum_variables; i++ ) {
      printf("\nVariable %d: ", i);
      print_Variable(i);
      printf("\n");
      for ( j = 0; j < gvariables[i].num_vals; j++) {
	print_Variable_Value(i, j, FALSE);
	printf("\n");
      }
      printf("Goal: ");
      if ( ggoal_on_var[i] == -1 ) {
	printf(" (no goal on this)");
      } else {
	print_Variable_Value(i, ggoal_on_var[i], FALSE);
      }
      printf("\n");
    }
    
    printf("\n\n========================OPs using Variables:\n");
    for ( op = 0; op < gnum_op_conn; op++ ) {
      if ( gop_conn[op].num_E == 0 ) {
	continue;
      }
      print_op_name(op);
      printf(": ");
      for ( i = 0; i < gop_conn[op].num_pre; i++ ) {
	print_Variable_Value(gop_conn[op].pre[i].var, gop_conn[op].pre[i].val, TRUE);
	if ( i < gop_conn[op].num_pre - 1 ) {
	  printf(", ");
	}
      }
      printf(" --> ");
      for ( i = 0; i < gop_conn[op].num_eff; i++ ) {
	print_Variable_Value(gop_conn[op].eff[i].var, gop_conn[op].eff[i].val, TRUE);
	if ( i < gop_conn[op].num_eff - 1 ) {
	  printf(", ");
	}
      }
      printf("\n");
    }
    
    printf("\n\n========================SUPPORT GRAPH:\n");
    print_SG();
    
    printf("\n\n========================DOMAIN TRANSITION GRAPHS:\n");
    for ( i = 0; i < gnum_variables; i++ ) {
      printf("\nVariable %d: ", i);
      print_Variable(i);
      printf("\n");
      print_DTG(i);
    }
  }




  /* we will always immediately close the output file, so that in case
   * the analysis fails somewhere we still have all the other data!
   */
  if ( gcmd_line.output_file ) {
    if ( (goutput_file = fopen( "TorchLight.txt", "w")) == NULL ) {
      printf("\nCan't open TorchLight.txt output file!\n\n");
      exit(1);
    }
    fprintf(goutput_file, "!!!!!-----------------------------------------------------------------\n");
    fprintf(goutput_file, "Input domain             : %s\n", gops_file);
    fprintf(goutput_file, "Input problem instance   : %s\n", gfct_file);
    fprintf(goutput_file, "----------------------------------------------------------------------\n");
    fprintf(goutput_file, "Number of samples        : %d\n", gcmd_line.num_samples);
    fprintf(goutput_file, "FD runtime for vars      : %.2f\n", gFDvariables_time);
    fprintf(goutput_file, "Total preprocess runtime : %.2f\n", 
	    gtempl_time + greach_time + grelev_time + gconn_time + gSG_DTG_time);
    fprintf(goutput_file, "Static SG+DTG analysis rt: %.2f\n", 
	    gSG_DTG_static_analyze_time);
    fprintf(goutput_file, "----------------------------------------------------------------------\n");
    fclose(goutput_file);
  }






  /* now we can finally do some analysis!
   */



  if ( gcmd_line.do_gDG ) {
    times( &start );
    
    analyze_global();
    
    times( &end );
    TIME( ganalyze_global_time );
    
    if ( gsuccess ) {
      printf("\nTorchLight guaranteed global analysis:");
      printf("\nNo local minima under h+, exit distance bound %d.",
	     ged_bound);
    } else {
      printf("\nTorchLight guaranteed global analysis:");
      printf("\nFailed.");
    }
    if ( gcmd_line.do_diagnose_gDG_successrate ) {
      printf("\nPercentage of successful x0/t0 gDGs    : %6.2f%% (%d of %d)", 
	     ((float) ggDG_num_successes) / ((float) ggDG_num_graphs) * 100.0,
	     ggDG_num_successes,
	     ggDG_num_graphs);
    }
    fflush(stdout);
    
    /* output file format:
     * <name>: [Done?] <0/1>, [Runtime?] <float>, [Successes?] <%>, [Dead-ends?] <%>, [min-ed] <int>, [mean-ed] <float>, [max-ed] <int> 
     * where not sensible, replace by -1
     */
    if ( gcmd_line.output_file ) {
      if ( (goutput_file = fopen( "TorchLight.txt", "a")) == NULL ) {
	printf("\nCan't open TorchLight.txt output file!\n\n");
	exit(1);
      }
      if ( gcmd_line.do_diagnose_gDG_successrate ) {
	fprintf(goutput_file, "Global                   : 1, %6.2f, %6.2f,  -1   , %3d, %6.2f, %3d\n", 
		ganalyze_global_time,
		((float) ggDG_num_successes) / ((float) ggDG_num_graphs) * 100.0,
		gsuccess ? ged_bound : -1,
		gsuccess ? ((float) ged_bound) : -1,
		gsuccess ? ged_bound : -1);
      } else {
	fprintf(goutput_file, "Global                   : 1, %6.2f, %6.2f,  -1   , %3d, %6.2f, %3d\n", 
		ganalyze_global_time,
		gsuccess ? 100.0 : 0.0,
		gsuccess ? ged_bound : -1,
		gsuccess ? ((float) ged_bound) : -1,
		gsuccess ? ged_bound : -1);
      }
      fclose(goutput_file);
    }
  } else {
    if ( gcmd_line.output_file ) {
      if ( (goutput_file = fopen( "TorchLight.txt", "a")) == NULL ) {
	printf("\nCan't open TorchLight.txt output file!\n\n");
	exit(1);
      }
      fprintf(goutput_file, "Global                   : 0,  -1   ,  -1   ,  -1   ,  -1,  -1   ,  -1\n");
      fclose(goutput_file);
    }
  }



  if ( gcmd_line.do_lDG ) {
    /* now do lDG-based analysis of initial state.
     */
    times( &start );
    
    if ( gcmd_line.negative_diagnose_all ) {
      gdo_negative_diagnosis = TRUE;
    } else {
      gdo_negative_diagnosis = FALSE;
    }
    analyze_local_lDG( &ginitial_state );
    if ( !gsuccess && !gcmd_line.negative_diagnose_all ) {
      /* now analyze this state again, for diagnosis!
       */
      gdo_negative_diagnosis = TRUE;
      analyze_local_lDG( &ginitial_state );
    }
    
    times( &end );
    TIME( ganalyze_local_ini_lDG_time );

    if ( gsuccess ) {
      printf("\n\nTorchLight guaranteed local analysis of initial state:");
      printf("\nNo local minima under h+, exit distance bound: %d.",
	     ged_bound);
    } else {
      if ( gdead_end ) {
	printf("\n\nTorchLight guaranteed local analysis of initial state:");
	printf("\nIs a dead end.");
      } else {
	printf("\n\nTorchLight guaranteed local analysis of initial state:");
	printf("\nFailed.");
      }
    }
    fflush(stdout);
    
    if ( gcmd_line.output_file ) {
      if ( (goutput_file = fopen( "TorchLight.txt", "a")) == NULL ) {
	printf("\nCan't open TorchLight.txt output file!\n\n");
	exit(1);
      }
      fprintf(goutput_file, "Local-guaranteed(ini)    : 1, %6.2f, %6.2f, %6.2f, %3d, %6.2f, %3d\n", 
	      ganalyze_local_ini_lDG_time,
	      gsuccess ? 100.0 : 0.0,
	      gdead_end ? 100.0 : 0.0,
	      gsuccess ? ged_bound : -1,
	      gsuccess ? ((float) ged_bound) : -1,
	      gsuccess ? ged_bound : -1);
      fclose(goutput_file);
    }
  } else {
    if ( gcmd_line.output_file ) {
      if ( (goutput_file = fopen( "TorchLight.txt", "a")) == NULL ) {
	printf("\nCan't open TorchLight.txt output file!\n\n");
	exit(1);
      }
      fprintf(goutput_file, "Local-guaranteed(ini)    : 0,  -1   ,  -1   ,  -1   ,  -1,  -1   ,  -1\n");
      fclose(goutput_file);
    }
  }



  /* now oDGplus-based analysis of initial state.
   */
  times( &start );

  if ( gcmd_line.negative_diagnose_all ) {
    gdo_negative_diagnosis = TRUE;
  } else {
    gdo_negative_diagnosis = FALSE;
  }
  analyze_local_oDGplus( &ginitial_state );
  if ( !gsuccess && !gcmd_line.negative_diagnose_all ) {
    /* now analyze this state again, for diagnosis!
     */
    gdo_negative_diagnosis = TRUE;
    analyze_local_oDGplus( &ginitial_state );
  }

  times( &end );
  TIME( ganalyze_local_ini_oDGplus_time );

  if ( gsuccess ) {
    printf("\n\nTorchLight approximate local analysis of initial state:");
    printf("\nNo local minima under h+, exit distance bound: %d.",
	   ged_bound);
  } else {
    if ( gdead_end ) {
      printf("\n\nTorchLight approximate local analysis of initial state:");
      printf("\nIs a dead end.");
    } else {
      printf("\n\nTorchLight approximate local analysis of initial state:");
      printf("\nFailed.");
    }
  }
  fflush(stdout);

  if ( gcmd_line.output_file ) {
    if ( (goutput_file = fopen( "TorchLight.txt", "a")) == NULL ) {
      printf("\nCan't open TorchLight.txt output file!\n\n");
      exit(1);
    }
    fprintf(goutput_file, "Local-approximate(ini)   : 1, %6.2f, %6.2f, %6.2f, %3d, %6.2f, %3d\n", 
	    ganalyze_local_ini_oDGplus_time,
	    gsuccess ? 100.0 : 0.0,
	    gdead_end ? 100.0 : 0.0,
	    gsuccess ? ged_bound : -1,
	    gsuccess ? ((float) ged_bound) : -1,
	    gsuccess ? ged_bound : -1);
    fprintf(goutput_file, "----------------------------------------------------------------------\n");
    fclose(goutput_file);
  }






  if ( gcmd_line.num_samples > 0 ) {



    /* now sample states for local analysis. First generate an array of
     * the states, so that we can use the same states in lDG and oDG+.
     */
    if ( gcmd_line.display_info >= 1 ) {
      printf("\n\nTorchLight: sampling random states"); 
      fflush(stdout);
    }
    times( &start );
    
    sample_states();
    
    times( &end );
    TIME( gsampling_time );
    if ( gcmd_line.display_info >= 1 ) {
      printf(" ... done.");
      fflush(stdout);
    }

    if ( gcmd_line.output_file ) {
      if ( (goutput_file = fopen( "TorchLight.txt", "a")) == NULL ) {
	printf("\nCan't open TorchLight.txt output file!\n\n");
	exit(1);
      }
      fprintf(goutput_file, "Sample states runtime    : %.2f\n", gsampling_time);
      fprintf(goutput_file, "----------------------------------------------------------------------\n");
      fclose(goutput_file);
    }




    if ( gcmd_line.do_lDG ) {
      times( &start );
      
      analyze_samples_local_lDG();
      
      times( &end );
      TIME( ganalyze_local_sample_lDG_time );
      
      printf("\n\nTorchLight guaranteed local analysis of sampled states:");
      printf("\nSuccess and hence no local minima under h+: %6.2f%%", 
	     gsuccess_percentage);
      printf("\nDead-end states: %6.2f%%", gdead_end_percentage);
      printf("\nExit distance bound min: %4d, mean: %6.2f, max: %4d",
	     gmin_ed_bound, gmean_ed_bound, gmax_ed_bound);
      fflush(stdout);
      
      if ( gcmd_line.output_file ) {
	if ( (goutput_file = fopen( "TorchLight.txt", "a")) == NULL ) {
	  printf("\nCan't open TorchLight.txt output file!\n\n");
	  exit(1);
	}
	fprintf(goutput_file, "Local-guaranteed(sample) : 1, %6.2f, %6.2f, %6.2f, %3d, %6.2f, %3d\n", 
		ganalyze_local_sample_lDG_time,
		gsuccess_percentage,
		gdead_end_percentage,
		gsuccess_percentage > 0 ? gmin_ed_bound : -1,
		gsuccess_percentage > 0 ? gmean_ed_bound : -1,
		gsuccess_percentage > 0 ? gmax_ed_bound : -1);
	fclose(goutput_file);
      }
    } else {
      if ( gcmd_line.output_file ) {
	if ( (goutput_file = fopen( "TorchLight.txt", "a")) == NULL ) {
	  printf("\nCan't open TorchLight.txt output file!\n\n");
	  exit(1);
	}
	fprintf(goutput_file, "Local-guaranteed(sample) : 0,  -1   ,  -1   ,  -1   ,  -1,  -1   ,  -1\n");
 	fclose(goutput_file);
      }
    }

    
   
    /* now analyze the samples using oDG+.
     */
    times( &start );
    
    analyze_samples_local_oDGplus();
     
    times( &end );
    TIME( ganalyze_local_sample_oDGplus_time );
    oDGsuccessrate = gsuccess_percentage;
   
    printf("\n\nTorchLight approximate local analysis of sampled states:");
    printf("\nSuccess and hence no local minima under h+: %6.2f%%", 
	   gsuccess_percentage);
    printf("\nDead-end states: %6.2f%%", gdead_end_percentage);
    printf("\nExit distance bound min: %4d, mean: %6.2f, max: %4d",
	   gmin_ed_bound, gmean_ed_bound, gmax_ed_bound);
    fflush(stdout);
    
    if ( gcmd_line.output_file ) {
      if ( (goutput_file = fopen( "TorchLight.txt", "a")) == NULL ) {
	printf("\nCan't open TorchLight.txt output file!\n\n");
	exit(1);
      }
      fprintf(goutput_file, "Local-approximate(sample): 1, %6.2f, %6.2f, %6.2f, %3d, %6.2f, %3d\n", 
	      ganalyze_local_sample_oDGplus_time,
	      gsuccess_percentage,
	      gdead_end_percentage,
	      gsuccess_percentage > 0 ? gmin_ed_bound : -1,
	      gsuccess_percentage > 0 ? gmean_ed_bound : -1,
	      gsuccess_percentage > 0 ? gmax_ed_bound : -1);
      fprintf(goutput_file, "----------------------------------------------------------------------\n");
      fclose(goutput_file);
    }

    


    /* now analyze the samples using local-EHC on monotone paths, if
     * wanted!
     */
    if ( gcmd_line.run_EHCanalyze ) {
      if ( gcmd_line.display_info >= 1 ) {
	printf("\n\n\nTorchLight: running local-monotone Enforced Hill-Climbing analysis... "); 
	fflush(stdout);
      }
      
      times( &start );
      
      analyze_samples_local_monotoneEHC();
      
      times( &end );
      TIME( ganalyze_local_sample_monotoneEHC_time );
      
      printf("\n\nTorchLight (approximate) monotone-local-EHC analysis of sampled states:");
      printf("\nSuccess and hence no local minima under h+: %6.2f%%", 
	     gsuccess_percentage);
      printf("\nDead-end states: %6.2f%%", gdead_end_percentage);
      printf("\nExit distance bound min: %4d, mean: %6.2f, max: %4d",
	     gmin_ed_bound, gmean_ed_bound, gmax_ed_bound);
      fflush(stdout);
      
      if ( gcmd_line.output_file ) {
	if ( (goutput_file = fopen( "TorchLight.txt", "a")) == NULL ) {
	  printf("\nCan't open TorchLight.txt output file!\n\n");
	  exit(1);
	}
	fprintf(goutput_file, "Local-monotoneEHC(sample): 1, %6.2f, %6.2f, %6.2f, %3d, %6.2f, %3d\n", 
		ganalyze_local_sample_monotoneEHC_time,
		gsuccess_percentage,
		gdead_end_percentage,
		gsuccess_percentage > 0 ? gmin_ed_bound : -1,
		gsuccess_percentage > 0 ? gmean_ed_bound : -1,
		gsuccess_percentage > 0 ? gmax_ed_bound : -1);
	fprintf(goutput_file, "----------------------------------------------------------------------\n");
	fclose(goutput_file);
      }
    } else {
      if ( gcmd_line.output_file ) {
	if ( (goutput_file = fopen( "TorchLight.txt", "a")) == NULL ) {
	  printf("\nCan't open TorchLight.txt output file!\n\n");
	  exit(1);
	}
	fprintf(goutput_file, "Local-monotoneEHC(sample): 0,  -1   ,  -1   ,  -1   ,  -1,  -1   ,  -1\n");
	fprintf(goutput_file, "----------------------------------------------------------------------\n");
	fclose(goutput_file);
      }
    }
  }



  /* for the moment, suppress only the diagnosis output, not the
   * diagnosis itself...
   */
  if ( gcmd_line.do_diagnose ) {
    /* now give diagnosis output!
     */
    printf("\n\nTorchLight DIAGNOSIS -- of SUCCESS:");
    
    if ( gcmd_line.do_diagnose_gDG_successrate ) {
      /* output info for gDGs only if some -- not all -- of the gDGs
       * were successful!
       */
      if ( ggDG_num_successes > 0 &&
	   ggDG_num_successes < ggDG_num_graphs ) {
	printf("\n\nSuccessful leaf operators op0 (on variable x0) in partially successful static global analysis:");
	for ( i = 0; i < ggDG_num_responsible_op0var0s; i++ ) {
	  printf("\n"); 
	  printf("%s", goperators[ggDG_responsible_op0s[i]]->name);
	  printf(" (");
	  print_Variable(ggDG_responsible_var0s[i]);
	  printf(")");
	  if ( gcmd_line.show_case_weight ) {
	    printf(": %6.2f%% of cases", 
		   ((float) ggDG_responsible_op0var0s_weights[i]) / ((float) ggDG_num_successes) * 100.0);
	  }
	}
      }
    }

    if ( gcmd_line.do_lDG &&
	 glDG_num_responsible_var0s > 0 ) {
      printf("\n\nSuccessful leaf variables x0 in guaranteed local analysis:");
      for ( i = 0; i < glDG_num_responsible_var0s; i++ ) {
	printf("\n"); 
	print_Variable(glDG_responsible_var0s[i]);
	if ( gcmd_line.show_case_weight ) {
	  printf(": %6.2f%% of cases", 
		 ((float) glDG_responsible_var0s_weights[i]) / ((float) glDG_num_successes) * 100.0);
	}
      }
    }
    
    if ( goDGplus_num_responsible_op0pred0s > 0 ) {
      printf("\n\nSuccessful leaf operators op0 (on variable x0) in approximate local analysis:");
      for ( i = 0; i < goDGplus_num_responsible_op0pred0s; i++ ) {
	printf("\n"); 
	printf("%s", goperators[goDGplus_responsible_op0s[i]]->name);
	printf(" (");
	printf("%s", gpredicates[goDGplus_responsible_pred0s[i]]);
	printf(")");
	if ( gcmd_line.show_case_weight ) {
	  printf(": %6.2f%% of cases", 
		 ((float) goDGplus_responsible_op0pred0s_weights[i]) / ((float) goDGplus_num_successes) * 100.0);
	}
      }
    }
    
    printf("\n\nTorchLight DIAGNOSIS -- of FAILURE:");
    
    if ( goDGplus_num_cycle_vars > 0 ) {
      printf("\n\nVariables on oDG+ cycles in approximate local analysis:");
      for ( i = 0; i < goDGplus_num_cycle_vars; i++ ) {
	printf("\n"); 
	print_Variable(goDGplus_cycle_vars[i]);
	if ( gcmd_line.show_case_weight ) {
	  printf(": %6.2f%% of cases", 
		 ((float) goDGplus_cycle_vars_weights[i]) / ((float) goDGplus_num_failures) * 100.0);
	}
      }
    }
    
    if ( goDGplus_num_nonleafbadtranss > 0 ) {
      printf("\n\nSide effects of invertible/induced transitions in approximate local analysis:");
      for ( i = 0; i < goDGplus_num_nonleafbadtranss; i++ ) {
	printf("\n"); 
	printf("%s", goperators[goDGplus_nonleafbadtrans_rops[i]]->name);
	printf(" (");
	print_Variable(goDGplus_nonleafbadtrans_seffvars[i]);
	printf(")");
	if ( gcmd_line.show_case_weight ) {
	  printf(": %6.2f%% of cases", 
		 ((float) goDGplus_nonleafbadtranss_weights[i]) / ((float) goDGplus_num_failures) * 100.0);
	}
      }
    }
      
      
    if ( gOLD_oDGplus_num_nonrecovered_RFC_intersects > 0 ) {
      printf("\n\nTop weighted non-recovered op0/variable in approximate local analysis:");
      gotone = TRUE;
      while ( gotone ) {
	gotone = FALSE;	
	maxw = -1;
	maxi = -1;
	for ( i = 0; i < gOLD_oDGplus_num_nonrecovered_RFC_intersects; i++ ) {
	  if ( maxw == -1 || gOLD_oDGplus_nonrecovered_RFC_intersects_weights[i] > maxw ) {
	    maxw = gOLD_oDGplus_nonrecovered_RFC_intersects_weights[i];
	    maxi = i;
	  }
	}	
	if ( maxi != -1 ) {
	  gotone = TRUE;
	  printf("\n");
	  if ( gcmd_line.show_case_weight ) {
	    printf("%6.2f%% of weight -- ",
		   ((float) gOLD_oDGplus_nonrecovered_RFC_intersects_weights[maxi]) /
		   ((float) gOLD_oDGplus_nonrecovered_RFC_intersects_totalweight) * 100.0);
	  }
	  printf("%s", goperators[gOLD_oDGplus_nonrecovered_RFC_intersect_op0s[maxi]]->name);
	  printf(" (");
	  print_Variable(gOLD_oDGplus_nonrecovered_RFC_intersect_vars[maxi]);
	  printf(")");

	  for ( j = maxi; j < gOLD_oDGplus_num_nonrecovered_RFC_intersects-1; j++ ) {
	    gOLD_oDGplus_nonrecovered_RFC_intersects_weights[j] = gOLD_oDGplus_nonrecovered_RFC_intersects_weights[j+1];
	    gOLD_oDGplus_nonrecovered_RFC_intersect_op0s[j] = gOLD_oDGplus_nonrecovered_RFC_intersect_op0s[j+1];
	    gOLD_oDGplus_nonrecovered_RFC_intersect_vars[j] = gOLD_oDGplus_nonrecovered_RFC_intersect_vars[j+1];
	  }
	  gOLD_oDGplus_num_nonrecovered_RFC_intersects--;
	}
      } 
    }
    

    if ( goDGplus_num_nonrecovered_RFC_intersects > 0 ) {
      printf("\n\nTop weighted non-recovered op/predicate in approximate local analysis:");
      gotone = TRUE;
      while ( gotone ) {
	gotone = FALSE;
	maxw = -1;
	maxi = -1;
	for ( i = 0; i < goDGplus_num_nonrecovered_RFC_intersects; i++ ) {	  
	  if ( maxw == -1 || goDGplus_nonrecovered_RFC_intersects_weights[i] > maxw ) {
	    maxw = goDGplus_nonrecovered_RFC_intersects_weights[i];
	    maxi = i;
	  }
	}
	if ( maxi != -1 ) {
	  gotone = TRUE;
	  printf("\n");
	  if ( gcmd_line.show_case_weight ) {
	    printf("%6.2f%% of weight -- ",
		   ((float) goDGplus_nonrecovered_RFC_intersects_weights[maxi]) /
		   ((float) goDGplus_nonrecovered_RFC_intersects_totalweight) * 100.0);
	  }
	  printf("%s", goperators[goDGplus_nonrecovered_RFC_intersect_op0s[maxi]]->name);
	  printf(" (");
	  printf("%s", gpredicates[goDGplus_nonrecovered_RFC_intersect_preds[maxi]]);
	  printf(")");

	  for ( j = maxi; j < goDGplus_num_nonrecovered_RFC_intersects-1; j++ ) {
	    goDGplus_nonrecovered_RFC_intersects_weights[j] = goDGplus_nonrecovered_RFC_intersects_weights[j+1];
	    goDGplus_nonrecovered_RFC_intersect_op0s[j] = goDGplus_nonrecovered_RFC_intersect_op0s[j+1];
	    goDGplus_nonrecovered_RFC_intersect_preds[j] = goDGplus_nonrecovered_RFC_intersect_preds[j+1];
	  }
	  goDGplus_num_nonrecovered_RFC_intersects--;
	}
      } 
    } 
          
  } /* endif do diagnose */







  if ( gcmd_line.run_FF ) {
    /* finally, if desired: run EHC and record its success/it's maximal
     * search depth!
     */
    if ( gcmd_line.display_info >= 1 ) {
      printf("\n\n\nTorchLight: running Enforced Hill-Climbing... "); 
      fflush(stdout);
    }
    times( &start );
    
    for ( i = 0; i < MAX_PLAN_LENGTH + 1; i++ ) {
      make_state( &(gplan_states[i]), gnum_ft_conn );
      gplan_states[i].max_F = gnum_ft_conn;
    }

    gEHC_max_search_depth = -1;
    doing_EHC = TRUE;

    if ( !do_enforced_hill_climbing( &ginitial_state, &ggoal_state ) ) {
      gEHC_success = 0;
    } else {
      gEHC_success = 1;
    }

    times( &end );
    TIME( gEHC_time );

    if ( gcmd_line.output_file ) {
      if ( (goutput_file = fopen( "TorchLight.txt", "a")) == NULL ) {
	printf("\nCan't open TorchLight.txt output file!\n\n");
	exit(1);
      }
      fprintf(goutput_file, "EHC runtime              : %.2f\n", gEHC_time);
      fprintf(goutput_file, "EHC success              : %d\n", gEHC_success);
      fprintf(goutput_file, "EHC max search depth     : %d\n", gEHC_max_search_depth);
      fprintf(goutput_file, "----------------------------------------------------------------------\n");
      fclose(goutput_file);
    }
  } else {/* endif run FF wanted */ 
    if ( gcmd_line.output_file ) {
      if ( (goutput_file = fopen( "TorchLight.txt", "a")) == NULL ) {
	printf("\nCan't open TorchLight.txt output file!\n\n");
	exit(1);
      }
      fprintf(goutput_file, "EHC runtime              : -1\n");
      fprintf(goutput_file, "EHC success              : -1\n");
      fprintf(goutput_file, "EHC max search depth     : -1\n");
      fprintf(goutput_file, "----------------------------------------------------------------------\n");
      fclose(goutput_file);
    }  
  }



  output_planner_info();



  printf("\n\n");
  exit( 0 );

}











/*
 *  ----------------------------- HELPING FUNCTIONS ----------------------------
 */












void output_planner_info( void )

{

  printf( "\n\nTime spent: %7.2f seconds instantiating %d easy, %d hard action templates", 
	  gtempl_time, gnum_easy_templates, gnum_hard_mixed_operators );
  printf( "\n            %7.2f seconds reachability analysis, yielding %d facts and %d actions", 
	  greach_time, gnum_pp_facts, gnum_actions );
  printf( "\n            %7.2f seconds creating final representation with %d relevant facts", 
	  grelev_time, gnum_relevant_facts );
  printf( "\n            %7.2f seconds building connectivity graph",
	  gconn_time );
  printf( "\n            %7.2f seconds in FD translator generating variables", 
	  gFDvariables_time );
  printf( "\n            %7.2f seconds preparing support graph and DTGs", 
	  gSG_DTG_time );
  printf( "\n            %7.2f seconds statically analyzing support graph and DTGs", 
	  gSG_DTG_static_analyze_time );
  printf( "\n            %7.2f seconds in guaranteed global analysis", 
	  ganalyze_global_time );
  if ( gcmd_line.do_lDG ) {
    printf( "\n            %7.2f seconds in guaranteed local analysis of initial state", 
	    ganalyze_local_ini_lDG_time );
  }
  printf( "\n            %7.2f seconds in approximate local analysis of initial state", 
	  ganalyze_local_ini_oDGplus_time );
  if ( gcmd_line.num_samples > 0 ) {
    printf( "\n            %7.2f seconds sampling states", 
	    gsampling_time );
    if ( gcmd_line.do_lDG ) {
      printf( "\n            %7.2f seconds in guaranteed local analysis of sample states", 
	      ganalyze_local_sample_lDG_time );
    }
    printf( "\n            %7.2f seconds in approximate local analysis of sample states", 
	    ganalyze_local_sample_oDGplus_time );
    if ( gcmd_line.run_EHCanalyze ) {
      printf( "\n            %7.2f seconds in local-monotone Enforced Hill-Climbing analysis of sample states", 
	      ganalyze_local_sample_monotoneEHC_time );
    }
  }
  printf( "\nTL TOTAL TIME: %7.2f seconds, of which %7.2f were spent computing relaxed plans for analysis.",
	  gtempl_time + greach_time + grelev_time + gconn_time + gFDvariables_time + gSG_DTG_time + gSG_DTG_static_analyze_time + ganalyze_global_time + ganalyze_local_ini_lDG_time + ganalyze_local_ini_oDGplus_time + ganalyze_local_sample_lDG_time + ganalyze_local_sample_oDGplus_time, grelaxed_plan_time );

  if ( gcmd_line.run_FF ) {
    printf( "\nEHC          : %7.2f seconds, success %d, max search depth %d.",
	    gEHC_time, gEHC_success, gEHC_max_search_depth);

  }  

}



void ff_usage( void )

{

  printf("\nUsage of TorchLight:\n");

  printf("\nOPTIONS   DESCRIPTIONS\n\n");
  printf("-p <str>    path for operator and fact file\n");
  printf("-o <str>    operator file name\n");
  printf("-f <str>    fact file name\n\n");

  printf("-s <num>    number of states to be sampled (preset: %d)\n",
	 gcmd_line.num_samples); 
  printf("-d <num>    depth of samples (*hFF(ini)) (preset: %d)\n\n",
	 gcmd_line.depth_samples); 

  printf("-G          do NOT do guaranteed global analysis\n");
  printf("-L          do NOT do guaranteed local analysis\n");
  printf("-R          do NOT do SG-root special case handling in Dcost\n");
  printf("-O          optimize over all choices of op0, x0 in oDG+ (default: stop at first success)\n");
  printf("-U          prune useless var0, ie var0 not in goal/op-pre\n");
  printf("-P <num>    level of P+ Replacement strategy (preset: %d)\n",
	 gcmd_line.replacement_level); 
  printf("      0     no P+ replacement\n");
  printf("      1     replace P+_>0 for reduction of R\n");
  printf("      2     + replace op0 for reduction of R\n");
  printf("      3     + replace for recovery of RFC facts\n");
  printf("-S          do not do recoverer-only-relevant-side-effects analysis\n\n");

  printf("-D          do diagnosis\n");
/*   printf("-G          do not compute and diagnose gDG success rate\n"); */
  printf("-V          do NOT prune useless var0, ie var0 not in goal/op-pre, in diagnosis\n");
  printf("-I          in RFC intersect diagnosis, do not try to invert op0\n");
  printf("-X          in RFC intersect diagnosis, do not ignore exchanged vars\n");
  printf("-2          diagnose failure separately, by 2nd analysis run only for non-success states\n");
  printf("-W          do not show case weight in diagnosis\n\n");

  printf("-E          use single-EHC iteration as local analysis technique\n");
  printf("-Z          run ONLY single-EHC iteration as local analysis technique\n\n");
  printf("-Q          run global EHC and provide data on success, max search depth\n");
  printf("-Y          run ONLY global EHC and provide data on success, max search depth\n");

  printf("-F          produce output data File\n");
  printf("-i <num>    run-time information level (preset: %d)\n",
	 gcmd_line.display_info);
  printf("      0     nothing\n");
  printf("      1     basic output and processing infos\n");
  printf("      2     + detailed analysis infos\n");
  printf("   > 100    various FF debugging infos\n\n");

  printf("-r <num>    random seed (preset: %d)\n\n",
	 gcmd_line.random_seed);
  

  if ( 0 ) {
    printf("-i <num>    run-time information level( preset: 1 )\n");
    printf("      0     only times\n");
    printf("      1     problem name, planning process infos\n");
    printf("    101     parsed problem data\n");
    printf("    102     cleaned up ADL problem\n");
    printf("    103     collected string tables\n");
    printf("    104     encoded domain\n");
    printf("    105     predicates inertia info\n");
    printf("    106     splitted initial state\n");
    printf("    107     domain with Wff s normalized\n");
    printf("    108     domain with NOT conds translated\n");
    printf("    109     splitted domain\n");
    printf("    110     cleaned up easy domain\n");
    printf("    111     unaries encoded easy domain\n");
    printf("    112     effects multiplied easy domain\n");
    printf("    113     inertia removed easy domain\n");
    printf("    114     easy action templates\n");
    printf("    115     cleaned up hard domain representation\n");
    printf("    116     mixed hard domain representation\n");
    printf("    117     final hard domain representation\n");
    printf("    118     reachability analysis results\n");
    printf("    119     facts selected as relevant\n");
    printf("    120     final domain and problem representations\n");
    printf("    121     connectivity graph\n");
    printf("    122     fixpoint result on each evaluated state\n");
    printf("    123     1P extracted on each evaluated state\n");
    printf("    124     H set collected for each evaluated state\n");
    printf("    125     False sets of goals <GAM>\n");
    printf("    126     detected ordering constraints leq_h <GAM>\n");
    printf("    127     the Goal Agenda <GAM>\n");

  }

}



Bool process_command_line( int argc, char *argv[] )

{

  char option;

  while ( --argc && ++argv ) {
    if ( *argv[0] != '-' || strlen(*argv) != 2 ) {
      return FALSE;
    }
    option = *++argv[0];

    switch ( option ) {
    case 'Q':
      gcmd_line.run_FF = TRUE;
      break;
    case 'Y':
      gcmd_line.run_only_FF = TRUE;
      break;
    case 'Z':
      gcmd_line.run_only_EHCanalyze = TRUE;
      break;
    case 'E':
      gcmd_line.run_EHCanalyze = TRUE;
      break;
    case 'F':
      gcmd_line.output_file = TRUE;
      break;
    case 'R':
      gcmd_line.do_SG_root = FALSE;
      break;
    case 'S':
      gcmd_line.do_recoverer_only_relevant = FALSE;
      break;
    case 'L':
      gcmd_line.do_lDG = FALSE;
      break;
    case 'G':
      gcmd_line.do_gDG = FALSE;
      break;
    case 'I':
      gcmd_line.do_diagnose_invertop0 = FALSE;
      break;
    case 'X':
      gcmd_line.do_diagnose_ignore_exchangedvars = FALSE;
      break;
    case 'U':
      gcmd_line.prune_useless_var0 = TRUE;
      break;
    case 'V':
      gcmd_line.diagnose_prune_useless_var0 = FALSE;
      break;
    case 'D':
      gcmd_line.do_diagnose = TRUE;
      break;
    case '2':
      gcmd_line.negative_diagnose_all = FALSE;
      break;
    case 'O':
      gcmd_line.optimize_over_op0var0 = TRUE;
      break;
/*     case 'G': */
/*       gcmd_line.do_diagnose_gDG_successrate = FALSE; */
/*       break; */
    case 'W':
      gcmd_line.show_case_weight = FALSE;
      break;
    default:
      if ( --argc && ++argv ) {
	switch ( option ) {
	case 'p':
	  strncpy( gcmd_line.path, *argv, MAX_LENGTH );
	  break;
	case 'o':
	  strncpy( gcmd_line.ops_file_name, *argv, MAX_LENGTH );
	  break;
	case 'f':
	  strncpy( gcmd_line.fct_file_name, *argv, MAX_LENGTH );
	  break;
	case 'i':
	  sscanf( *argv, "%d", &gcmd_line.display_info );
	  break;
	case 'd':
	  sscanf( *argv, "%d", &gcmd_line.depth_samples );
	  break;
	case 's':
	  sscanf( *argv, "%d", &gcmd_line.num_samples );
	  break;
	case 'r':
	  sscanf( *argv, "%d", &gcmd_line.random_seed );
	  break;
	case 'P':
	  sscanf( *argv, "%d", &gcmd_line.replacement_level );
	  break;
	default:
	  printf( "\nff: unknown option: %c entered\n\n", option );
	  return FALSE;
	}
      } else {
	return FALSE;
      }
    }
  }

  if ( gcmd_line.replacement_level < 0 || gcmd_line.replacement_level > 3 ) {
    printf("\nUnknown replacement level %d", gcmd_line.replacement_level);
    return FALSE;
  }

  return TRUE;

}

